#include "parser.h"
#include "common.h" 
 
 
///////////////////////////////////////////////////////////////////////////////


const uint8_t CCParser::startUserDataHeader[]	= {0x00, 0x00, 0x01, 0xB2};
const uint8_t CCParser::endUserDataHeader[]	  = {0x00, 0x00, 0x00, 0x01};
const uint8_t CCParser::dishDVBHeader[]		    = {0x05, 0x02}; 
const uint8_t CCParser::r4000DVBHeader[]	    = {0xBB, 0x02};
const uint8_t CCParser::r5000DVBHeader[]	    = {0x99, 0x02};
const uint8_t CCParser::dvdHeader[]		        = {0x43, 0x43};

#define MAX_RETRIES 20
 
///////////////////////////////////////////////////////////////////////////////

 
CCParser::CCParser(CCManager* m, Interpreter* i) : cThread("CC Parser")
{
  dprint(L_DBG, "Parser  : Constructor.");
  manager = m;
  interpreter = i;
  
  inBuffer  = new cRingBufferLinear(KILOBYTE(256), 64, false, "CC Parser In");
  
  unknownUserData = 0;
}


//-----------------------------------------------------------------------------

CCParser::~CCParser()
{
  Cancel(3);
  
  delete inBuffer; inBuffer = NULL;

  dprint(L_DBG, "Parser  : Destructor.");
}


//-----------------------------------------------------------------------------

void CCParser::Action(void) 
{
  dprint(L_DBG, "Parser  : Thread started.");
  int noStartFound=0;
  int noDelete=0;
  
  while (Running())
  {
    //uint64_t st = cTimeMs::Now();
    //uint64_t extraSleep = 0;
    if (!inBuffer) break;
    
    int length=0;
    uint8_t* data = inBuffer->Get(length);
    if (data && length != 0)
    {
      //dprint(L_DBG, "Parser  : Got %d bytes from input buffer.", length);
      int used = 0;
      
      uint8_t* d = data;
      
      while (d - data < length-4)
      {
        uint8_t* start = NULL;
        while (d - data < length-4) {
          if (Equals(d, startUserDataHeader, 4)) {
            start = d;
            d+=4;
            break;
          }
          d++;  
        }
        if (!start) // Did not find start header
        {
          noStartFound++;
          //dprint(L_DBG, "Parser  : No start header found (%d)", noStartFound);
          used = length;
          break;
        }
      
        uint8_t* end = NULL;
        while (d - data < length-4) {
          if (Equals(d, endUserDataHeader, 4)) {
            end = d;
            d+=4;
            break;
          }  
          d++;
        }
        if (!end)
        {
          // This seems to cause a problem...
          used = start - data;
          dprint(L_DBG, "Parser  : No end found, used %d of %d", used, length);          
          break;
        }
        
        // Found both start and end
#if 0                         
          fprintf(stderr,"\n****************************************\n");
          for (uint8_t* dq = start; dq<end+4; dq++)
          fprintf(stderr,"%02X ", dq[0]);
          fprintf(stderr,"\n****************************************\n\n");
#endif           
        ProcessUserData(start+4, end-start-4-1);
        noStartFound = 0; 
      }
    
      //dprint(L_DBG, "Parser  : Deleting %d of %d bytes from input buffer.", used, length);
      if (used == 0) {
        noDelete++;      
        dprint(L_DBG, "Parser  : Deleting 0 of %d bytes from input buffer. (%d)", length, noDelete);

        if (noDelete == MAX_RETRIES) {  
          dprint(L_ERR, "Parser  : %d bytes of stale data in buffer. Dumping.", length);
          used = length;
          noDelete=0;
        }    
      }  
      inBuffer->Del(used);
      
    }
    else {
      noStartFound++;
      dprint(L_DBG, "Parser  : No data in input buffer (%d)", noStartFound);
      //extraSleep = 500;
    }
    
    if (noStartFound == MAX_RETRIES) {
      manager->NoDataFound();
    }  
       
    //cCondWait::SleepMs(SLEEP_TIME - cTimeMs::Now() + st + extraSleep);
    cCondWait::SleepMs(SLEEP_TIME);
  }
  
  
  dprint(L_DBG, "Parser  : Thread ended.");
}


//-----------------------------------------------------------------------------

void CCParser::ProcessUserData(uint8_t* data, int length)
{
  if (length < 2) return;
  //dprint(L_DBG, "Parser  : Processing %d bytes of user data.", length);

  if (Equals(data, dishDVBHeader, 2)) {
    ProcessDishCC(data+2, length-2);
    unknownUserData = 0; 
  }  
  else if (Equals(data, r4000DVBHeader, 2)) {
    dprint(L_MSG, "Parser  : Unsuported CC Type Detected.");
    unknownUserData++;
  }
  else if (Equals(data, r5000DVBHeader, 2)) {
    dprint(L_MSG, "Parser  : Unsuported CC Type Detected.");
    unknownUserData++;
  }  
  else if (Equals(data, dvdHeader, 2)) {  
    dprint(L_MSG, "Parser  : Unsuported CC Type Detected.");
    unknownUserData++;
  }
  else {
    dprint(L_MSG, "Parser  : Unknown user data header {%02X, %02X}", data[0], data[1]);
    unknownUserData++;
  }
  
  if (unknownUserData == MAX_RETRIES) {
    manager->NoDataFound();
  }  
}


//-----------------------------------------------------------------------------

static uint8_t fBuff[64];
static int fPos = 0;

void CCParser::ProcessDishCC(uint8_t* data, int length)
{
/*  
  fprintf(stderr,"\n========================================\n");
  for (int i=0; i<length; i++)
   fprintf(stderr,"%02X ", data[i]&0x7F);
  fprintf(stderr,"\n========================================\n\n"); 
*/
  uint8_t* d = data+5;
  
  uint8_t type = d[0];
  
  if (type == 0x02 && d+2-data < length)
  {
    Store(d+2, 2);
    
    if( (d[4] == 0x04) && ((d[2] & 0x7F) < 0x20) )
      Store(d+2, 2);
    
    //d+=4;  
  }
  
  else if (type == 0x04 && d+5-data < length)
  {
    Store(d+2, 4);
  }
  
  else if (type == 0x05 && d+10-data < length)
  {
    if (fPos>0) Store(fBuff, fPos);
      //for (int i=0; i<fPos; i++)
      //  fprintf(stderr, "%c", fBuff[i]&0x7F);
    
    fPos = 0;

    fBuff[fPos++] = d[9];
    fBuff[fPos++] = d[10];
    
    if(d[7] == 0x02)
    {
      if( (d[11] == 0x04) && ((d[9] & 0x7F) < 0x20) ) {
        fBuff[fPos++] = d[9];
        fBuff[fPos++] = d[10];
      }
    }
    else
    {	
      //Store(d+11, 2);	
      fBuff[fPos++] = d[11];
      fBuff[fPos++] = d[12];				
		} 
		    
    d+=13;
  }
  else { fprintf(stderr, "UNKNOWN CASE: 0x%02X at index %d of %d\n", type, d-data, length);  }
 
}

void CCParser::Store(uint8_t* data, int length)
{
  interpreter->Put(data, length);
}

//-----------------------------------------------------------------------------

void CCParser::Put(uint8_t* data, int length)
{
  int p = inBuffer->Put(data, length);
  if (p != length) { 
    dprint(L_ERR, "Parser  : Input buffer overflow (%d bytes dropped)", length - p); 
    inBuffer->ReportOverflow(length - p);
  }
    
  // dprint(L_DBG, "Parser  : Stored %d of %d bytes", p, length);   
}


///////////////////////////////////////////////////////////////////////////////

